use std::path::PathBuf;
use std::io::ErrorKind;
use std::fs::{self, DirEntry};

use regex::Regex;
use clap::Parser;

const START_KEY: &str = "#=mgen_start=#\n";
const END_KEY: &str = "#=mgen_end=#";

#[derive(Parser, Debug)]
struct Args {
    /// Replace default make output with nice build messages (not yet implemented)
    #[arg(short, long)]
    pretty: bool,

    /// Path to makefile
    #[arg(short, long, default_value = "./Makefile")]
    makefile: String,
}

#[derive(Debug)]
#[allow(dead_code)]
enum BuildRuleKind<T> {
    Object(T),
    Target(T),
    Directory(T),
    Convenience(T),
}

#[derive(Debug,Clone)]
struct BuildRule {
    target: String,
    prerequisites: Vec<String>,
    recipe_commands: Vec<String>,
}


#[derive(Debug)]
#[allow(dead_code)]
struct Makefile {
    path: String,
    prologue: String,
    mgen_zone: String,
    epilogue: String,
    include_dirs: Vec<String>,
}

#[derive(Debug)]
struct RuleList(Vec<BuildRuleKind<BuildRule>>);

impl BuildRule {
    fn get_headers_from_c_file(path: &str, makefile: &Makefile) -> Vec<String> {
        let mut headers: Vec<String> = Vec::new();
        let file_contents = fs::read_to_string(path).unwrap_or_else(|e| {
            panic!("error: could not read file {path}: {e}");
        });

        let file_path = PathBuf::from(path);
        let file_dir = file_path.parent().unwrap();

        let includes_regex = Regex::new(r#"#include\s+"(.*?)""#).unwrap();
        let caps = includes_regex.captures_iter(&file_contents);
        for (_, [include_file]) in caps.map(|c| c.extract()) {
            match file_dir.join(PathBuf::from(include_file)).canonicalize() {
                Ok(path) => {
                    headers.push(String::from(path.to_str().unwrap()));
                    continue;
                },
                Err(_) => (),
            }

            let mut done = false;
            for dir in makefile.include_dirs.iter() {
                let dir = PathBuf::from(&dir);
                match dir.join(PathBuf::from(include_file)).canonicalize() {
                    Ok(path) => {
                        headers.push(String::from(path.to_str().unwrap()));
                        done = true;
                        break;
                    },
                    Err(_) => (),
                }
            }
            if done {
                continue;
            }
            let path = file_dir.join(PathBuf::from(include_file));
            headers.push(String::from(path.to_str().unwrap()));
        }

        headers
    }
    fn new_from_c_file(path: &str, makefile: &Makefile) -> BuildRule {

        let mut prerequisites: Vec<String> = Vec::new();

        prerequisites.push(String::from(path));

        prerequisites.append(&mut BuildRule::get_headers_from_c_file(path, makefile));

        prerequisites.push(String::from("| $(OBJDIR)"));

        let file_path = PathBuf::from(path);

        let obj_name = file_path.with_extension("o");
        let obj_name = obj_name.file_name().unwrap();

        let obj_path = PathBuf::from("$(OBJDIR)").join(obj_name);
        let obj_path = obj_path.to_str().unwrap();

        let compile_command = format!("$(CC) $(CFLAGS) $(INCS) -o {} -c {}", obj_path, path);

        BuildRule { target: String::from(obj_path), prerequisites, recipe_commands: vec![compile_command] }
    }
}

impl ToString for BuildRule {
    fn to_string(&self) -> String {
        let mut ret = String::new();
        ret.push_str(&format!("{}:", self.target));
        for prereq in &self.prerequisites {
            ret.push_str(&format!(" {}", prereq));
        }
        ret.push_str(&format!("\n"));
        for recipe_comm in &self.recipe_commands {
            ret.push_str(&format!("\t{}\n", recipe_comm));
        }
        ret
    }
}

impl Makefile {
    fn new_from_path(path: &str) -> Makefile {
        let contents = match fs::read_to_string(path) {
            Ok(v) => v,
            Err(e) => match e.kind() {
                // doesn't matter if the file doesn't exist, it will be created 
                // later.
                ErrorKind::NotFound => return Makefile { 
                    path: String::from(path), prologue: String::from(""),
                    mgen_zone: String::from(""), epilogue: String::from(""),
                    include_dirs: vec![String::from("")] },
                _ => panic!("error: cannot read makefile: {e}")
            }
        };

        let contains_start = contents.find(START_KEY).is_some();
        let contains_end = contents.find(END_KEY).is_some();

        if (contains_start && !contains_end) || (!contains_start && 
                                                                contains_end) {
            panic!("error: makefile contains one mgen guard but not the other. cannot continue.");
        }

        if ! contains_start && contains_end {
            let include_dirs = Makefile::get_include_dirs(&contents);
            return Makefile { path: String::from(path), prologue: contents,
                mgen_zone: String::from(""), epilogue: String::from(""), include_dirs };
        }

        // contains the part before the start key and everything after
        let pre_and_after: Vec<&str> = contents.split(START_KEY).collect();
        // contains the space between start and end keys and also the space after
        let gen_and_post: Vec<&str> = pre_and_after[1].split(END_KEY).collect();

        let include_dirs = Makefile::get_include_dirs(&contents);
        
        Makefile { 
            path: String::from(path),
            prologue: String::from(pre_and_after[0]),
            mgen_zone: String::from(gen_and_post[0]),
            epilogue: String::from(gen_and_post[1]),
            include_dirs
        }
    }
    
    fn get_include_dirs(search: &String) -> Vec<String> {
        let makefile_regex = Regex::new(r"^INCS\s*=\s*(.*?)$").unwrap();
        let caps = makefile_regex.captures_iter(search);
        let (_, [dirlist]) = caps.map(|c| c.extract()).last().unwrap();

        let mut include_dirs: Vec<String> = Vec::new();

        let incdir_regex = Regex::new(r"-s*-I\s*(.*?)\s*").unwrap();
        let caps = incdir_regex.captures_iter(&dirlist);
        for (_, [dir]) in caps.map(|c| c.extract()) {
            include_dirs.push(String::from(dir));
        }

        include_dirs
    }
        
    fn contains_builddir_def(&self) -> bool {
        let re = Regex::new(r"\s+BUILDDIR = ").unwrap();
        re.is_match(&(self.prologue)) || re.is_match(&(self.epilogue))
    }

    fn contains_objdir_def(&self) -> bool {
        let re = Regex::new(r"\s+OBJDIR = ").unwrap();
        re.is_match(&(self.prologue)) || re.is_match(&(self.epilogue))
    }
     
    fn write_new_generated_content(&self, generated_content: String) -> std::io::Result<()> {
        let content = [&self.prologue, START_KEY, &generated_content, END_KEY, &self.epilogue].concat();

        fs::write(&self.path, &content)?;

        Ok(())
    }
}

fn join_build_rule_vec(vector: &Vec<BuildRule>, seperator: &str) -> String{
    let mut ret = String::new();

    for (i, rule) in vector.iter().enumerate() {
        if i != 0 {
            ret.push_str(seperator);
        }
        ret.push_str(&rule.to_string());
    }

    ret
}

impl ToString for RuleList {
    fn to_string(&self) -> String {
        let targets = join_build_rule_vec(&self.get_target_rules(), "");
        let objects = join_build_rule_vec(&self.get_object_rules(), "");
        let directories = join_build_rule_vec(&self.get_directory_rules(), "");
        let conveniences = join_build_rule_vec(&self.get_convenience_rules(), "");

        [targets, objects, directories, conveniences].concat()
    }
}

impl RuleList {
    fn new() -> RuleList{
        RuleList(vec![])
    }

    fn recursively_gen_objects(&mut self, proj_dir_path: &str, makefile: &Makefile) -> std::io::Result<()>{
        let files = fs::read_dir(proj_dir_path)?;
        for file in files {
            let file = file?;
            let is_dir = file.metadata()?.is_dir();

            // ignore hidden files
            if file_is_hidden(&file) {
                continue;
            }

            let path = file.path();

            if is_dir {
                self.recursively_gen_objects(path.to_str().unwrap(), makefile)?;
            } else {
                match path.extension() {
                    Some(v) => {
                        let v = v.to_str().unwrap();
                        if v != "c" {
                            continue;
                        }
                        let path = path.to_str().unwrap();
                        self.0.push(BuildRuleKind::Object(BuildRule::new_from_c_file(path, makefile)));
                    },
                    None => {
                        continue;
                    },
                }
            }
        }
        Ok(())
    }

    fn get_obj_targets(&self) -> Vec<String>{
        let mut ret: Vec<String> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Object(r) => ret.push(r.target.clone()),
                _ => break,
            }
        }
        ret
    }

    fn add_target_rule(&mut self) {
        let mut objects = self.get_obj_targets();
        let target_comm = String::from(
            format!("$(CC) $(LDFLAGS) $(INCS) -o $(BUILDDIR)/$(TARGET) {} $(LDLIBS)", objects.join(" "))
        );

        objects.push(String::from("| $(OBJDIR) $(BUILDDIR)"));
        let prereqs = objects;

        self.0.push(BuildRuleKind::Target(BuildRule {
            target: String::from("$(BUILDDIR)/$(TARGET)"),
            prerequisites: prereqs,
            recipe_commands: vec![target_comm],
        }))
    }

    fn add_directory_rule(&mut self, dir: &str) {
        self.0.push(BuildRuleKind::Directory(BuildRule {
            target: dir.to_string(),
            prerequisites: vec![],
            recipe_commands: vec![format!("mkdir -p {dir}").to_string()]
        }));
    }

    fn add_clean_rule(&mut self) {
        self.0.push(BuildRuleKind::Convenience(BuildRule {
            target: String::from(".PHONY: clean\nclean"),
            prerequisites: vec![],
            recipe_commands: vec![String::from("rm -r $(OBJDIR)"), String::from("rm -r $(BUILDDIR)")]
        }));
    }

    fn add_run_rule(&mut self) {
        self.0.push(BuildRuleKind::Convenience(BuildRule {
            target: String::from(".PHONY: run\nrun"),
            prerequisites: vec![String::from("$(BUILDDIR)/$(TARGET)")],
            recipe_commands: vec![String::from("./$(BUILDDIR)/$(TARGET)")]
        }));
    }

    fn get_target_rules(&self) -> Vec<BuildRule> {
        let mut ret: Vec<BuildRule> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Target(r) => ret.push(r.clone()),
                _ => continue,
            }
        }
        ret
    }

    fn get_object_rules(&self) -> Vec<BuildRule> {
        let mut ret: Vec<BuildRule> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Object(r) => ret.push(r.clone()),
                _ => continue,
            }
        }
        ret
    }

    fn get_directory_rules(&self) -> Vec<BuildRule> {
        let mut ret: Vec<BuildRule> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Directory(r) => ret.push(r.clone()),
                _ => continue,
            }
        }
        ret
    }

    fn get_convenience_rules(&self) -> Vec<BuildRule> {
        let mut ret: Vec<BuildRule> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Convenience(r) => ret.push(r.clone()),
                _ => continue,
            }
        }
        ret
    }
}

fn file_is_hidden(file: &DirEntry) -> bool{
    let file_name = file.file_name();
    let str_name = file_name.to_str().unwrap();
    let first_char = str_name.chars().nth(0).unwrap();
    if first_char == '.' {
        true
    } else {
        false
    }
}

fn main() {
    let args = Args::parse();
    let makefile = Makefile::new_from_path(&args.makefile);

    let mut generated_content = String::new();

    // Only include directory defaults if the user has not specified them
    if ! makefile.contains_builddir_def() {
        generated_content.push_str("BUILDDIR = ./build\n");
    }

    if ! makefile.contains_objdir_def() {
        generated_content.push_str("OBJDIR = $(BUILDDIR)/obj\n");
    }

    let mut rule_list = RuleList::new();
    rule_list.recursively_gen_objects(".", &makefile).unwrap_or_else(|e| {
        panic!("error: problem in recursive search: {e}");
    });

    rule_list.add_target_rule();

    rule_list.add_directory_rule("$(BUILDDIR)");
    rule_list.add_directory_rule("$(OBJDIR)");
    rule_list.add_clean_rule();
    rule_list.add_run_rule();

    generated_content.push_str(&rule_list.to_string());

    // Write makefile
    makefile.write_new_generated_content(generated_content).unwrap_or_else(|e| {
        panic!("error: could not write makefile at {}: {e}", makefile.path);
    });
}
